signature ABSTRACT_SYNTAX =
sig
    structure Support : SUPPORT
    structure Types : TYPES
	where type support = Support.support

    type var = Var.var
    type location = Location.location

    type support = Support.support
    type tp = Types.tp
    type ltp = Types.ltp

    datatype binop = PLUS | MINUS | MULT | EQL
	
    datatype term =
	Empty 
      | Print of term
      |	Nat of int
      | True 
      | False 
      | Nil
      | Cons of term * term
      | Car of term 
      | Cdr of term
      | Var of var
      | BinOp of binop * term * term
      | If of term * term * term 
      | Fn of var * term
      | Fix of var * var * term
      | App of term * term 
      | Pair of term * term 
      | Fst of term 
      | Snd of term 
      | Box of term 
      | Dia of clos
      | Let of decl list * term 
      | Choose of term
      | Nu of var * term
      | ExpSubst of (var * term) list * term
      | Raise of var * term
      | Handle of (var * var * term) list * term	
      | Shift of var * var * term
      | Reset of var * term
      | OfTp of term * tp 

    and decl = 
	NameDecl of var * tp
      |	ValDecl of var * term 
      | BoxDecl of var * term 
      | DiaDecl of var * term

    and lterm =
	LEmpty of location
      | LPrint of lterm * location
      |	LNat of int * location
      | LTrue of location
      | LFalse of location
      | LNil of location
      | LCons of lterm * lterm * location
      | LCar of lterm * location
      | LCdr of lterm * location
      | LVar of var * location
      | LBinOp of binop * lterm * lterm * location
      | LIf of lterm * lterm * lterm * location
      | LFn of var * lterm * location
      | LFix of var * var * lterm * location
      | LApp of lterm * lterm * location
      | LPair of lterm * lterm * location
      | LFst of lterm * location
      | LSnd of lterm * location
      | LBox of lterm * location
      | LDia of lclos * location
      | LLet of ldecl list * lterm * location
      | LChoose of lterm * location
      | LNu of var * lterm * location
      | LExpSubst of (var * lterm * location) list * lterm * location
      | LRaise of var * lterm * location
      | LHandle of (var * var * lterm * location) list * lterm * location
      | LShift of var * var * lterm * location
      | LReset of var * lterm * location
      | LOfTp of lterm * ltp * location
 
    and ldecl = 
	LNameDecl of var * ltp * location
      | LValDecl of var * lterm * location
      | LBoxDecl of var * lterm * location
      | LDiaDecl of var * lterm * location

    (* types for possibility *)

    and clos =
        Clos of (var * term) list * term 
      | CLet of decl list * clos 

    and lclos =
        LClos of (var * lterm * location) list * lterm * location
      | LCLet of ldecl list * lclos * location



    val termToString : term -> string
    val toTerm : lterm -> term
    val termLoc : lterm -> location

    val declToString : decl -> string
    val toDecl : ldecl -> decl
    val declLoc : ldecl -> location

    val closToString : clos -> string
    val toClos : lclos -> clos
    val closLoc : lclos -> location

    type subst = term Var.VarMap.map

    val substTerm : (subst * term) -> term
    val substDecls : (subst * decl list) -> (subst * decl list)
    val substType : subst * tp -> tp
    val substSupport : subst * support -> support 
    val substClos : (subst * clos) -> clos

    val termEq : term * term -> bool
    val closEq : clos * clos -> bool

end


