functor ObjectTypes(structure Support : SUPPORT) :> 
TYPES where type support = Support.support
=
struct
    type var = Var.var
    type location = Location.location
    type support = Support.support
	
    datatype tp = 
	UNIT
      |	NAT 
      | BOOL
      | NATLIST
      | VAR of var
      | PROD of tp * tp
      | ARROW of tp * tp 
      | NARROW of tp * tp
      | BOX of support * tp
      | DIA of support * tp

    datatype ltp = 
	LUNIT of location
      |	LNAT  of location
      | LBOOL of location
      | LNATLIST of location
      | LVAR of var * location
      | LPROD of ltp * ltp * location
      | LARROW of ltp * ltp * location
      | LNARROW of ltp * ltp * location
      | LBOX of support * ltp * location
      | LDIA of support * ltp * location


    fun toString (tp:tp):string =
	(case tp of
	     UNIT => "unit"
           | NAT => "int"
	   | BOOL => "bool"
	   | NATLIST => "intlist"
	   | VAR x => Var.toString x
	   | PROD (tp1, tp2)=> 
		 let val s1 = toString tp1
		     val s2 = toString tp2
		 in
		     s1^" * "^s2
		 end
	   | ARROW (tp1, tp2) => 
		 let val s1 = toString tp1
		     val s2 = toString tp2
		 in
		     s1^" -> "^s2
		 end

	   | NARROW (tp1, tp2) => 
		 let val s1 = toString tp1
		     val s2 = toString tp2
		 in
		     s1^" -x> "^s2
		 end
	   | BOX (C, tp) => 
		 let val s1 = Support.toString C
		     val s2 = toString tp
		 in
		     "["^s1^"]("^s2^")"
		 end
	   | DIA (C, tp) => 
		 let val s1 = Support.toString C
		     val s2 = toString tp
		 in
		     "<"^s1^">"^s2
		 end)


    fun fromLtp (ltp:ltp):tp =
	(case ltp of
	     LUNIT _ => UNIT
	   | LNAT _ => NAT
	   | LBOOL _ => BOOL
	   | LNATLIST _ => NATLIST
	   | LVAR (x, _) => VAR x
	   | LPROD (ltp1, ltp2, _) => 
		 let val tp1 = fromLtp ltp1
		     val tp2 = fromLtp ltp2
		 in 
		     PROD (tp1, tp2)
		 end
	   | LARROW (ltp1, ltp2, _) =>
		 let val tp1 = fromLtp ltp1
		     val tp2 = fromLtp ltp2
		 in 
		     ARROW (tp1, tp2)
		 end
	   | LNARROW  (ltp1, ltp2, _) =>
		 let val tp1 = fromLtp ltp1
		     val tp2 = fromLtp ltp2
		 in 
		     NARROW (tp1, tp2)
		 end
	   | LBOX (C, ltp1, _) => 
		 let val tp1 = fromLtp ltp1
		 in
		     BOX (C, tp1)
		 end
	   | LDIA (C, ltp1, _) => 
		 let val tp1 = fromLtp ltp1
		 in
		     DIA (C, tp1)
		 end)


    fun getLoc (ltp:ltp) =
	(case ltp of (
	     LUNIT loc
	   | LNAT loc 
	   | LBOOL loc
	   | LNATLIST loc
	   | LVAR (_, loc)
	   | LPROD (_, _, loc)
	   | LARROW (_, _, loc) 
	   | LNARROW  (_, _, loc) 
	   | LBOX (_, _, loc)
	   | LDIA (_, _, loc)) => loc)


    fun equal (tp1, tp2) = 
	(case (tp1, tp2) of
	     (UNIT, UNIT) => true
	   | (NAT, NAT) => true
	   | (BOOL, BOOL) => true
	   | (NATLIST, NATLIST) => true
	   | (PROD (A1, A2), PROD (B1, B2)) =>
		 equal (A1, B1) andalso equal (A2, B2)
	   | (ARROW (A1, A2), ARROW (B1, B2)) =>
		 equal (A1, B1) andalso equal (A2, B2)
	   | (NARROW (A1, A2), NARROW (B1, B2)) =>
		 equal (A1, B1) andalso equal (A2, B2)
	   |  (BOX (C1, A1), BOX (C2, A2)) =>
		 Support.equal (C1, C2) andalso equal (A1, A2)
	   |  (DIA (C1, A1), DIA (C2, A2)) =>
		 Support.equal (C1, C2) andalso equal (A1, A2)
	   |  _ => false)

    fun inType (v : Var.var, tp) = 
	(case tp of
	     (UNIT | NAT | BOOL | NATLIST | VAR _) => false
	   | (PROD (tp1, tp2) | ARROW (tp1, tp2) | NARROW (tp1, tp2)) => 
		 inType (v, tp1) orelse inType (v, tp2)
	   | (BOX (C, tp1) | DIA (C, tp1)) => 
		 Support.isElement (v, C) orelse inType (v, tp1))

end
