functor Typechecking (structure Absyn : ABSTRACT_SYNTAX 
                      structure Ctx : CONTEXT
                          where Absyn = Absyn
	              val lang : Language.lang) :> 
    TYPECHECK where Absyn = Absyn
and Ctx = Ctx
=

struct

    structure Absyn = Absyn
    structure Ctx = Ctx

    open Absyn
    open Types
    open Ctx


    fun inferTerm (ctx, lt, C) : tp = 
	(case lt of 
	     LEmpty _ => UNIT
	   | LPrint (lt', _) => 
		 let val _ = inferTerm (ctx, lt', C)
		 in
		     UNIT
		 end
	   | LNat _ => NAT
	   | LTrue _ => BOOL
	   | LFalse _ => BOOL
	   | LNil _ => NATLIST
	   | LCons (lt1, lt2, loc) => 
		 (if checkTerm (ctx, lt1, NAT, C) then
		      if checkTerm (ctx, lt2, NATLIST, C) then NATLIST
		      else 
			  let val inexp = termToString (toTerm lt)
			  in
			      Error.error ("tail of an intlist is not of intlist type", 
					   loc, SOME inexp)
			  end
		  else
		      let val inexp = termToString (toTerm lt)
		      in
			  Error.error ("head of an intlist is not of int type", 
				       loc, SOME inexp)
		      end)
	      
	   | LCar (lt1, loc) => 
		 (if (checkTerm (ctx, lt1, NATLIST, C)) then NAT
		  else
		      let val inexp = termToString (toTerm lt)
		      in
			  Error.error ("operand is not an intlist", loc, SOME (inexp))
		      end)

	   | LCdr (lt1, loc) => 
		 (if (checkTerm (ctx, lt1, NATLIST, C)) then NATLIST
		  else
		      let val inexp = termToString (toTerm lt)
		      in
			  Error.error ("operand is not an intlist", loc, SOME (inexp))
		      end)

	   | LVar (x, loc) => 
		 (case (Ctx.find (ctx, x)) of
		      NONE => Error.error ("unknown variable or name "^(Var.toString x), loc, NONE) : tp
		    | SOME (Variable (B, D)) => 
			  if (Support.leq (D, C)) then B
			  else
			      let val info1 = "variable support: ["^(Support.toString D)^"]"
				  val info2 = "expected support: ["^(Support.toString C)^"]"
			      in
				  Error.error ("variable "^(Var.toString x)^
					       " has unexpected support", 
					       loc, SOME(info1^"\n"^info2))
			      end
		    | SOME (Name B) => 
			  (case lang of
			       (Language.DBIND | Language.STATE) =>
				   if (Support.agrees (x, C)) then B
				   else 
				       let val info = "support: ["^(Support.toString C)^"]"
				       in
					   Error.error ("name "^(Var.toString x)^
							" used, but not in the support", loc, SOME(info))
				       end
			     | (Language.EXCPTN | Language.CONTS) => 
				       Error.error ("name "^(Var.toString x)^
						    " used as a variable", loc, NONE)))

	   | LBinOp ((PLUS|MINUS|MULT), e1, e2, loc) => 
  	         if (checkTerm (ctx, e1, NAT, C) 
		      andalso checkTerm (ctx, e2, NAT, C))
		 then NAT 
		 else 
		      let val inexp = termToString (toTerm lt)
		      in
			  Error.error ("operator and operand don't agree", loc, SOME inexp)
		      end

	   | LBinOp (EQL, e1, e2, loc) =>
		 (let val d = 
		      (checkTerm (ctx, e1, NAT, C) andalso
		       checkTerm (ctx, e2, NAT, C)) 
		      handle Error.Error _ => 
			  (checkTerm (ctx, e1, NATLIST, C) andalso
			   checkTerm (ctx, e2, NATLIST, C))
		  in
		      if d then BOOL
		      else
			  let val inexp = termToString (toTerm lt)
			  in
			      Error.error ("operator and operand don't agree", loc, SOME inexp)
			  end
		  end)

	   | LApp (e1, e2, loc) => 
		 let val A = inferTerm (ctx, e1, C)
		 in 
		     case A of 
			 (ARROW (A1, B)) =>
			     if checkTerm (ctx, e2, A1, C) then B
			     else
				 let val inexp = termToString (toTerm lt)
				 in
				     Error.error ("operator and operand don't agree", loc, SOME inexp)
				 end
		       | _ => 
                             let val inexp = termToString (toTerm lt)
				 val info = "operator type: " ^ (Types.toString A)
			     in
				 Error.error ("operator is not a function", 
					      loc, SOME (info ^"\n"^ "in expression:\n"^inexp))
			     end
		 end

	   | LFst (e', loc) => 
                 let val A = inferTerm (ctx, e', C)
		 in
		     case A of
			 (PROD (A1, A2)) => A1
		       | _ => 
			     let val inexp = termToString (toTerm lt)
				 val info = "operand type: " ^ (Types.toString A)
			     in
				 Error.error ("operand is not a pair", 
					      loc, SOME (info ^"\n"^ "in expression:\n"^inexp))
			     end
		 end

	   | LSnd (e', loc) =>
                 let val A = inferTerm (ctx, e', C)
		 in
		     case A of
			 (PROD (A1, A2)) => A2
		       | _ => 
			     let val inexp = termToString (toTerm lt)
				 val info = "operand type: " ^ (Types.toString A)
			     in
				 Error.error ("operand is not a pair", 
					      loc, SOME (info ^"\n"^ "in expression:\n"^inexp))
			     end
		 end

	   | LChoose (e', loc) => 
		 let val A = inferTerm (ctx, e', C)
		 in
		     case A of
			 NARROW (A1, A2) => A2
		       | _ =>
			     let val inexp = termToString (toTerm lt)
				 val info = "operand type: " ^ (Types.toString A)
			     in
				 Error.error ("operand is not an irrelevant abstraction", 
					      loc, SOME (info ^"\n"^"in expression:\n"^inexp))
			     end
		 end
	     
	   (* begin section: admissible inference derivations *)
       
	   | LIf (e1, e2, e3, loc) => 
		 if checkTerm (ctx, e1, BOOL, C)
		     then 
			 let val A2 = inferTerm (ctx, e2, C)
			     val A3 = inferTerm (ctx, e3, C)
			 in
			     if Types.equal (A2, A3) then A2
			     else 
				 let val inexp = termToString (toTerm lt)
				     val info1 = "first branch: " ^(Types.toString A2)
				     val info2 = "second branch: "^(Types.toString A3)
				 in
				     Error.error ("types of branches don't agree", 
						  loc, SOME (info1^ "\n" ^info2^ "\n" ^"in expression:\n" ^ inexp))
				 end
			 end
		 else 
		     let val inexp = termToString (toTerm lt)
		     in
			 Error.error ("condition is not boolean", loc, SOME ("\n"^"in expression:\n" ^inexp))
		     end

	   | LPair (e1, e2, loc) => PROD (inferTerm (ctx, e1, C), inferTerm (ctx, e2, C))

	   | LLet (nil, e2, loc) => Error.error ("parser error: empty declaration list", loc, NONE)

	   | LLet (decls, e2, loc) => 
		 inferLets (ctx, decls, e2, C)

	   | LExpSubst (subst, e', loc) => 
		 if (lang <> Language.DBIND) andalso
		     (lang <> Language.STATE) then 
		     Error.error ("syntax error: unknown construct", loc, NONE)
		 else
		     let val D = inferSubst (ctx, subst, C)
		     in
			 inferTerm (ctx, e', D)
		     end

	   | LHandle (handler, e', loc) => 
		 if lang <> Language.EXCPTN then 
		     Error.error ("syntax error: unknown construct", loc, NONE)
		 else
		     let val D = inferHandlerSupport (ctx, handler, C)
			 val A = inferTerm (ctx, e', D)
		     in
			 if checkHandler (ctx, handler, A, D, C) then A
			 else Error.error ("expression and handler types don't agree", 
					   loc, NONE)
		     end

	   | LReset (x, e', loc) => 
		 if lang <> Language.CONTS then 
		     Error.error ("syntax error: unknown construct", loc, NONE)
		 else
		     (case Ctx.find (ctx, x) of
			  SOME (Name A) => 
			      ((let val _ = checkTerm (ctx, e', A, Support.add (x, C))
				in A
				end) handle Error.Error _ => 
				    let val _ = checkTerm (ctx, e', A, Support.singleton x)
				    in
					A
				    end)
			| SOME (Variable _) => 
			      Error.error ("name "^(Var.toString x)^" used as a variable", 
					   loc, NONE)
			| NONE => Error.error ("unbound name "^(Var.toString x), loc, NONE))
			  
			  
	   (* end section: admissible inference derivations *)

	   | LOfTp (e', lt', loc) => 
		 let val _ = Ctx.okTp (ctx, lt')
		     val t = Types.fromLtp lt'
		 in
		     if checkTerm (ctx, e', t, C) then t
		     else
			 let val inexp = termToString (toTerm lt)
			 in
			     Error.error ("expression doesn't match constraint", loc, 
					  SOME ("\n"^"in expression: "^inexp))
			 end
		 end
           | _ => 
		 let val s = termToString (toTerm lt)
		 in
		     Error.error ("can't infer type (must check it)", termLoc (lt), 
				  SOME ("\n"^"in expression: "^s))
		 end)



    and inferSubst (ctx, subst, C) =
	(case subst of
	     nil => C
	   | ((x, e, loc)::subst') => 
		 (case (Ctx.find (ctx, x)) of
		      SOME (Name A) =>
			  if checkTerm (ctx, e, A, C) then 
			      let val D = inferSubst (ctx, subst', C)
			      in
				  Support.add (x, D)
			      end
			  else 
			      let val inexp = (Var.toString x)^" -> "^(termToString (toTerm e))
			      in
				  Error.error ("name and term in assigment don't agree", loc, 
					       SOME ("\n"^"in expression: "^inexp))
			      end
		    | _ => Error.error ("unbound name", loc, NONE)))


    and inferHandlerSupport (ctx, subst, C) =
	(case subst of
	     nil => C
	   | ((x, y, e, loc)::subst') => 
		 (case (Ctx.find (ctx, x)) of
		      SOME (Name A) =>
			  let val D = inferHandlerSupport (ctx, subst', C)
			  in
			      Support.add (x, D)
			  end
		    | _ => Error.error ("unbound name "^(Var.toString x), loc, NONE)))


(*
    and inferDecl (ctx, ldecl, C) : ctx * tp =
	(case ldecl of
	     LNameDecl (x, ltp, loc) =>
		 if Ctx.okTp (ctx, ltp) then
		     let val tp = Types.fromLtp ltp 
		     in 
			 (Ctx.insertName (ctx, x, tp), tp)
		     end
		 else
		     Error.error ("ill-formed type", loc, NONE)
	   | LValDecl (x, e, loc) =>
		 let val A = inferTerm (ctx, e, C)
		 in
		     (Ctx.insertVar (ctx, x, A, Support.empty), A)
		 end

	   | LBoxDecl (u, e, loc) => 
		 let val A = inferTerm (ctx, e, C)
		 in
		     case A of
			 BOX (D, A') => 
			     (Ctx.insertVar (ctx, u, A', D), A)
		       | _ => 
			     let val indecl = declToString (toDecl ldecl) 
				 val info = "expression type: "^(Types.toString A)
			     in
				 Error.error ("expression in box declaration is not of boxed type", 
					      loc, SOME (info^"\n"^"in declaration: "^indecl))
			     end
		 end)
*)

    and inferLets (ctx, ldecls, e, C) : tp =
	(case ldecls of
	     nil => inferTerm (ctx, e, C)
	   | (LNameDecl (x, ltp, loc)::lds') =>
		 let val _ = Ctx.okTp (ctx, ltp)
		 in
		     let val tp1 = Types.fromLtp ltp 
			 val x' = Var.derived_var x
			 val ctx' = Ctx.shadowName (ctx, x, x')
			 val ctx'' = Ctx.insertName (ctx', x, tp1)
			 val s = Var.VarMap.insert (Var.VarMap.empty, x, Var x')
			 val C' = substSupport (s, C)
			 val tp = inferLets (ctx'', lds', e, C')
		     in
			 if Types.inType (x, tp) then 
			     Error.error ("name "^(Var.toString x)^"escaping the scope", loc, NONE)
			 else tp
		     end
		 end
	   | (LValDecl (x, e', loc)::lds') =>
		 let val A = inferTerm (ctx, e', C)
		     val ctx' = Ctx.insertVar (ctx, x, A, Support.empty)
		 in
		     inferLets (ctx', lds', e, C)
		 end

	   | ((ldecl as LBoxDecl (u, e', loc))::lds') => 
		 let val A = inferTerm (ctx, e', C)
		 in
		     case A of
			 BOX (D, A') => 
			     let val ctx' = Ctx.insertVar (ctx, u, A', D)
			     in
				 inferLets (ctx', lds', e, C)
			     end
		       | _ => 
			     let val indecl = declToString (toDecl ldecl) 
				 val info = "expression type: "^(Types.toString A)
			     in
				 Error.error ("expression in box declaration is not of boxed type", 
					      loc, SOME (info^"\n"^"in declaration: "^indecl))
			     end
		 end)



    and inferClos (ctx, lf, C) : support * tp = 
	(case lf of
	     LCLet (cdecls, lf', _) => 
		 inferCLets (ctx, cdecls, lf', C)
(*
		 let val (ctx', D1) = inferCDecls (ctx, cdecls, C)
		 in
		     inferClos (ctx', lf, D1)
		 end
*)
	   | _ => 
		 let val s = closToString (toClos lf)
		 in
		     Error.error ("can't infer closure type (must check it)", closLoc (lf), 
				  SOME ("\n"^"in closure: "^s))
		 end)
	     

     
    and inferCLets (ctx, lcdecl, f, C) : support * tp =
	(case lcdecl of
	     nil => inferClos (ctx, f, C)
	   | (LNameDecl (x, ltp, loc)::ds) =>	 
		 let val _ = Ctx.okTp (ctx, ltp)
		 in
		     let val tp1 = Types.fromLtp ltp 
			 val x' = Var.derived_var x
			 val ctx' = Ctx.shadowName (ctx, x, x')
			 val ctx'' = Ctx.insertName (ctx', x, tp1)
			 val s = Var.VarMap.insert (Var.VarMap.empty, x, Var x')
			 val C' = substSupport (s, C)
			 val (sprt, tp) = inferCLets (ctx'', ds, f, C')
		     in
			 if (Types.inType (x, tp) orelse
			     Support.isElement (x, sprt)) then 
			     Error.error ("name "^(Var.toString x)^"escaping the scope", loc, NONE)
			 else (sprt, tp)
		     end
		 end

	   | (LValDecl (x, e', loc)::ds) => 
		 let val A = inferTerm (ctx, e', C)
		     val ctx' = Ctx.insertVar (ctx, x, A, Support.empty)
		 in
		     inferCLets (ctx', ds, f, C)
		 end
	     
	   | ((lcdecl as LBoxDecl (u, e', loc))::ds) => 
		 let val A = inferTerm (ctx, e', C)
		 in
		     case A of
			 BOX (D1, A') =>   
			     let val ctx' = Ctx.insertVar (ctx, u, A', D1)
			     in
				 inferCLets (ctx', ds, f, C)
			     end
		       | _ => 
			     let val indecl = declToString (toDecl lcdecl) 
				 val info = "expression type: "^(Types.toString A)
			     in
				 Error.error ("expression in box declaration is not of boxed type", 
					      loc, SOME (info^"\n"^"in declaration: "^indecl))
			     end
		 end
	   | ((lcdecl as LDiaDecl (x, e, loc))::ds) => 
		 let val A = inferTerm (ctx, e, C)
		 in
		     case A of
			 DIA (D1, A') => 
			     let val ctx' = Ctx.insertVar (ctx, x, A', Support.empty)
			     in
				 inferCLets (ctx', ds, f, D1)
			     end
		       | _ => 
			     let val indecl = declToString (toDecl lcdecl) 
				 val info = "expression type: "^(Types.toString A)
			     in
				 Error.error ("expression in dia declaration is not of diamond type", 
					      loc, SOME (info^"\n"^"in declaration: "^indecl))
			     end
		 end)



	     
    and checkTerm (ctx, lt, A, C) : bool =
	(case lt of
	     LPair (e1, e2, loc) => 
		 (case A of 
		      PROD (A1, A2) => 
			  checkTerm (ctx, e1, A1, C) andalso
			  checkTerm (ctx, e2, A2, C)
		    | _ => 
			  let val inexp = termToString (toTerm lt)
			      val info = "specified type: "^(Types.toString A)
			  in
			      Error.error ("expression doesn't check against the type", 
					   loc, SOME(info^"\n"^"in expression: "^inexp))
			  end)

	   | LFn (x, e', loc) => 
		 (case A of
		      ARROW (A1, A2) =>
			  let val ctx' = Ctx.insertVar (ctx, x, A1, Support.empty)
			  in
			      checkTerm (ctx', e', A2, Support.empty)
			  end
		    | _ => 
			  let val inexp = termToString (toTerm lt)
			      val info = "specified type: "^(Types.toString A)
			  in
			      Error.error ("function doesn't check against the type", 
					   loc, SOME (info^"\n"^"in expression: "^inexp))
			  end)

	   | LFix (f, x, e', loc) => 
		 (case A of
		      ARROW (A1, A2) =>
			  let val ctx' = Ctx.insertVar (ctx, f, A, Support.empty)
			      val ctx'' = Ctx.insertVar (ctx', x, A1, Support.empty)
			  in
			      checkTerm (ctx'', e', A2, Support.empty)
			  end
		    | _ => 
			  let val inexp = termToString (toTerm lt)
			      val info = "specified type: "^(Types.toString A)
			  in
			      Error.error ("function doesn't check against the type", 
					   loc, SOME (info^"\n"^"in expression: "^inexp))
			  end)

	   | LBox (e', loc) => 
	         (case A of
		      BOX (D, A') =>
			  checkTerm (ctx, e', A', D)
		    | _ =>
			  let val inexp = termToString (toTerm lt)
			      val info = "specified type: "^(Types.toString A)
			  in
			      Error.error ("boxed term must check against boxed type", 
					   loc, SOME(info^"\n"^"in expression: "^inexp))
			  end)

	   | LDia (f', loc) => 
		 if lang <> Language.STATE then 
		     Error.error ("syntax error: unknown construct", loc, NONE)
		 else
		     (case A of
			  DIA (D, A') =>
			      checkClos (ctx, f', A', D, C)
			| _ =>
			      let val inexp = termToString (toTerm lt)
				  val info = "specified type: "^(Types.toString A)
			      in
				  Error.error ("dia term must check against diamond type", 
					       loc, SOME(info^"\n"^"in expression: "^inexp))
			      end)


	   | LNu (X, e', loc) => 
		 (case A of
		      NARROW (A1, A2) =>
			  let val ctx' = Ctx.insertName (ctx, X, A1)
			  in
			      checkTerm (ctx', e', A2, Support.empty)
			  end
		    | _ =>
			  let val inexp = termToString (toTerm lt)
			      val info = "specified type: "^(Types.toString A)
			  in
			      Error.error ("irrelevant abstraction doesn't check against the type", 
					   loc, SOME(info^"\n"^"in expression: "^inexp))
			  end)
		      

	   | LRaise (x, e', loc) => 
		 if lang <> Language.EXCPTN then 
		     Error.error ("syntax error: unknown construct", loc, NONE)
		 else
		     (case (Ctx.find (ctx, x)) of
			  NONE => Error.error ("unknown exception name "^(Var.toString x), loc, NONE) 
			| SOME (Name B) => 
			      if (Support.agrees (x, C)) then 
				  checkTerm (ctx, e', B, C)
			      else 
				  let val inexp = termToString (toTerm lt)
				      val info = "support: ["^(Support.toString C)^"]"
				  in
				      Error.error ("raised exception "^(Var.toString x)^
						   " is not in the support", 
						   loc, SOME ("in expression: "^inexp^"\n"^info))
				  end
			      
			| SOME (Variable (B, D)) => 
				  let val inexp = termToString (toTerm lt)
				  in
				      Error.error ("cannot raise variable exception "^(Var.toString x), 
						   loc, SOME ("in expression: "^inexp))
				  end)
			  
	   | LShift (x, k, e', loc) => 
		 if lang <> Language.CONTS then 
		     Error.error ("syntax error: unknown construct", loc, NONE)
		 else
		     (case (Ctx.find (ctx, x)) of
			  NONE => Error.error ("unknown name "^(Var.toString x), loc, NONE) 
			| SOME (Name AA) => 
			      if Support.agrees (x, C) then 
				  let val C' = Support.remove (x, C)
				      val ctp = ARROW (BOX (C, A), BOX (C, AA))
				      val ctx' = Ctx.insertVar (ctx, k, ctp, Support.empty)
				  in
				      checkTerm (ctx', e', AA, C')
				  end
			      else 
				  let val inexp = termToString (toTerm lt)
				      val info1 = "shifting to label: "^(Var.toString x)
				      val info2 = "support: ["^(Support.toString C)^"]"
				  in
				      Error.error ("shift and support don't match", 
						   loc, SOME (info1^"\n"^info2^"\n"^
							      "in expression: "^inexp))
				  end
			  
			| SOME (Variable _) => 
				   let val inexp = termToString (toTerm lt)
				   in
				       Error.error ("attempt to shift to a non-name",
						    loc, SOME ("in expression: "^inexp))
				   end)
			  
			  
           (* begin section: checking of let forms *)

	   | LIf (e1, e2, e3, loc) => 
		 (checkTerm (ctx, e1, BOOL, C) andalso
		  checkTerm (ctx, e2, A, C) andalso 
		  checkTerm (ctx, e3, A, C))

	   | LLet (nil, e2, loc) => Error.error ("parser error: empty declaration list", loc, NONE)

	   | LLet (decls, e2, loc) => 
		 checkLets (ctx, decls, e2, A, C)

	   | LExpSubst (subst, e', loc) => 
		 if (lang <> Language.DBIND) andalso 
		     (lang <> Language.STATE) then 
		     Error.error ("syntax error: unknown construct", loc, NONE)
		 else
		     let val D = inferSubst (ctx, subst, C)
		     in
			 checkTerm (ctx, e', A, D)
		     end

	   | LHandle (handler, e', loc) => 
		 if lang <> Language.EXCPTN then 
		     Error.error ("syntax error: unknown construct", loc, NONE)
		 else
		     let val D = inferHandlerSupport (ctx, handler, C)
		     in
			 checkTerm (ctx, e', A, D) andalso
			 checkHandler (ctx, handler, A, D, C) 
		     end

	   | LReset (x, e', loc) => 
		 if lang <> Language.CONTS then 
		     Error.error ("syntax error: unknown construct", loc, NONE)
		 else
		     (case Ctx.find (ctx, x) of
			  SOME (Name AA) => 
			      Types.equal (AA, A) andalso
			      (checkTerm (ctx, e', A, Support.add(x, C))
			       handle Error.Error _ => 
				   checkTerm (ctx, e', A, Support.singleton x))
			| SOME (Variable _) => 
			      Error.error ("destination label "^(Var.toString x)^
					   " used as a variable", loc, NONE)
			| NONE => Error.error ("unbound name "^(Var.toString x), loc, NONE))



           (* end section: checking of let forms *)

	 | (LEmpty loc | LPrint (_, loc) | LNat (_, loc) | LTrue loc | LFalse loc | 
            LNil loc | LCons (_, _, loc) | LCar (_, loc) | LCdr (_, loc) | LVar (_, loc) | 
            LBinOp (_, _, _, loc) | LApp (_, _, loc) | LFst (_, loc) | LSnd (_, loc) |  
            LChoose (_, loc) | LOfTp (_, _, loc)) =>

 		 let val A' = inferTerm (ctx, lt, C)
		 in
		     if Types.equal (A, A') then true
		     else
			 let val inexp = termToString (toTerm lt)
			     val info1 = "expression type: "^(Types.toString A')
			     val info2 = "specified type: "^(Types.toString A)
			 in
			     Error.error ("expression doesn't check against the type", 
					  loc, SOME(info1^"\n"^info2^"\n"^"in expression: "^inexp))
			 end
		 end)


    and checkHandler (ctx, handler, B, C, D) =
	(case handler of 
	     nil => Support.leq (C, D)
	   | ((x, y, e, loc)::handler') => 
		 (case Ctx.find (ctx, x) of
		      SOME (Name A) => 
			  let val ctx' = Ctx.insertVar (ctx, y, A, Support.empty)
			  in 
			      checkTerm (ctx', e, B, D) andalso
			      checkHandler (ctx, handler', B, Support.remove (x, C), D)
			  end
		    | SOME (Variable _) => 
			  Error.error ("cannot handle variable expression "^(Var.toString x), loc, NONE)
		    | NONE => 
			  Error.error ("unbound name "^(Var.toString x), loc, NONE)))
		     





    and checkLets (ctx, ldecls, e, tp, C) : bool =
	(case ldecls of
	     nil => checkTerm (ctx, e, tp, C)
	   | (LNameDecl (x, ltp, loc)::lds') =>
		 let val _ = Ctx.okTp (ctx, ltp)
		 in
		     let val tp1 = Types.fromLtp ltp 
			 val x' = Var.derived_var x
			 val ctx' = Ctx.shadowName (ctx, x, x')
			 val ctx'' = Ctx.insertName (ctx', x, tp1)
			 val s = Var.VarMap.insert (Var.VarMap.empty, x, Var x')
			 val tp' = substType (s, tp)
			 val C' = substSupport (s, C)
		     in
			 checkLets (ctx'', lds', e, tp', C')
		     end
		 end
	   | (LValDecl (x, e', loc)::lds') =>
		 let val A = inferTerm (ctx, e', C)
		     val ctx' = Ctx.insertVar (ctx, x, A, Support.empty)
		 in
		     checkLets (ctx', lds', e, tp, C)
		 end

	   | ((ldecl as LBoxDecl (u, e', loc))::lds') => 
		 let val A = inferTerm (ctx, e', C)
		 in
		     case A of
			 BOX (D, A') => 
			     let val ctx' = Ctx.insertVar (ctx, u, A', D)
			     in
				 checkLets (ctx', lds', e, tp, C)
			     end
		       | _ => 
			     let val indecl = declToString (toDecl ldecl) 
				 val info = "expression type: "^(Types.toString A)
			     in
				 Error.error ("expression in box declaration is not of boxed type", 
					      loc, SOME (info^"\n"^"in declaration: "^indecl))
			     end
		 end)

	     

    and checkClos (ctx, f, A, D, C) : bool =
	(case f of
	     LClos (assgns, e, loc) =>
		 let val _ = checkTerm (ctx, e, A, D)
		     val D' = inferSubst (ctx, assgns, C)
		 in
		     if Support.leq (D, D') then true
		     else
			 let val inclos = closToString (toClos f)
			     val info1 = "maximal substitution domain: "^(Support.toString D')
			     val info2 = "specified substitution domain: "^(Support.toString D)
			 in
			     Error.error ("substitution in closure has smaller domain than specificied",
					  loc, SOME (info1^"\n"^info2^"\n"^
						     "in closure: "^inclos))
			 end
		 end
	     
	     
	   | LCLet (nil, f2, loc) => Error.error ("parser error: empty declaration list", loc, NONE)
		 
	   | LCLet (cdecls, f2, loc) => 
		 checkCLets (ctx, cdecls, f2, A, D, C))



     
    and checkCLets (ctx, lcdecl, f, A, D, C) : bool =
	(case lcdecl of
	     nil => checkClos (ctx, f, A, D, C)
	   | (LNameDecl (x, ltp, loc)::ds) =>	 
		 let val _ = Ctx.okTp (ctx, ltp)
		 in
		     let val tp1 = Types.fromLtp ltp 
			 val x' = Var.derived_var x
			 val ctx' = Ctx.shadowName (ctx, x, x')
			 val ctx'' = Ctx.insertName (ctx', x, tp1)
			 val s = Var.VarMap.insert (Var.VarMap.empty, x, Var x')
			 val A' = substType (s, A)
			 val C' = substSupport (s, C)
			 val D' = substSupport (s, D)
		     in
			 checkCLets (ctx'', ds, f, A', D', C')
		     end
		 end

	   | (LValDecl (x, e', loc)::ds) => 
		 let val A1 = inferTerm (ctx, e', C)
		     val ctx' = Ctx.insertVar (ctx, x, A1, Support.empty)
		 in
		     checkCLets (ctx', ds, f, A, D, C)
		 end
	     
	   | ((lcdecl as LBoxDecl (u, e', loc))::ds) => 
		 let val A1 = inferTerm (ctx, e', C)
		 in
		     case A1 of
			 BOX (D1, A') =>   
			     let val ctx' = Ctx.insertVar (ctx, u, A', D1)
			     in
				 checkCLets (ctx', ds, f, A, D, C)
			     end
		       | _ => 
			     let val indecl = declToString (toDecl lcdecl) 
				 val info = "expression type: "^(Types.toString A)
			     in
				 Error.error ("expression in box declaration is not of boxed type", 
					      loc, SOME (info^"\n"^"in declaration: "^indecl))
			     end
		 end
	   | ((lcdecl as LDiaDecl (x, e, loc))::ds) => 
		 let val A1 = inferTerm (ctx, e, C)
		 in
		     case A1 of
			 DIA (D1, A') => 
			     let val ctx' = Ctx.insertVar (ctx, x, A', Support.empty)
			     in
				 checkCLets (ctx', ds, f, A, D, D1)
			     end
		       | _ => 
			     let val indecl = declToString (toDecl lcdecl) 
				 val info = "expression type: "^(Types.toString A)
			     in
				 Error.error ("expression in dia declaration is not of diamond type", 
					      loc, SOME (info^"\n"^"in declaration: "^indecl))
			     end
		 end)



    fun typechecks ldecls =
	if lang = Language.STATE then
	    let val _ = checkCLets (Ctx.empty, ldecls, 
				    LClos (nil, LEmpty (Location.dummy), Location.dummy),
				    UNIT, Support.empty, Support.empty)
	    in true end
	else
	    let val _ = inferLets (Ctx.empty, ldecls, LEmpty (Location.dummy), Support.empty)
	    in
		true
	    end

	
(*
    fun fromFile (file : string) = 
	let fun inferDeclsPrint (ctx, decls) = 
	    (case decls of 
		 nil => (ctx, "")
	       | (d::ds) => 
		     let val (ctx', A) = inferDecl (ctx, d, Support.empty)
			 val s1 = declToString (toDecl d)
			 val s2 = Types.toString A
			 val (ctx'', s3) = inferDeclsPrint (ctx', ds)
		     in
			 (ctx'', s1 ^ " : " ^ s2 ^ "\n" ^ s3)
		     end)
	    val decls = Parse.fromFile file
	    val ss = (#2 (inferDeclsPrint (Var.VarMap.empty, decls)))
		     handle TypeError msg => msg^"\n"
		
	in 
	    print ss
	end
*)
end
