structure Tokens = Tokens
 
type pos = int
type svalue = Tokens.svalue
type ('a,'b) token = ('a,'b) Tokens.token
type lexresult= (svalue,pos) token

(* position is line numbers *) 
val pos = LineNumber.lineNum

fun eof () = Tokens.EOF(!pos,!pos)
fun error (e,l : int,_) = TextIO.output (TextIO.stdOut, String.concat[
        "line ", (Int.toString l), ": ", e, "\n"
      ])
 
(* val rm_quotes = String.translate (fn c => if c = #"\"" then ""  (*"*)
                                          else String.str c)
*)

val comment_counter = ref 0

%%
%header (functor NUMODLexFun(structure Tokens: NUMOD_TOKENS));
%s COMMENT;

alpha=[A-Za-z];
digit=[0-9];
ws = [\ \t];
%%

<INITIAL>"\n"       => (pos := (!pos) + 1; lex());
<INITIAL>{ws}+    => (lex());
<INITIAL>{digit}+ => (Tokens.NUM(valOf (Int.fromString yytext), !pos, !pos));
<INITIAL>"true"     => (Tokens.TRUE(!pos,!pos));
<INITIAL>"false"    => (Tokens.FALSE(!pos,!pos));
<INITIAL>"fun"      => (Tokens.FUN(!pos,!pos));
<INITIAL>"fn"       => (Tokens.FN(!pos,!pos));
<INITIAL>"."        => (Tokens.DOT(!pos,!pos));
<INITIAL>","        => (Tokens.COMMA(!pos,!pos));
<INITIAL>":"        => (Tokens.COLON(!pos,!pos));
<INITIAL>"if"       => (Tokens.IF(!pos,!pos));
<INITIAL>"then"       => (Tokens.THEN(!pos,!pos));
<INITIAL>"else"       => (Tokens.ELSE(!pos,!pos));
<INITIAL>"let"       => (Tokens.LET(!pos,!pos));
<INITIAL>"box"       => (Tokens.BOX(!pos,!pos));
<INITIAL>"dia"       => (Tokens.DIA(!pos,!pos));
<INITIAL>"val"       => (Tokens.VAL(!pos,!pos));
<INITIAL>"in"       => (Tokens.IN(!pos,!pos));
<INITIAL>"end"       => (Tokens.END(!pos,!pos));
<INITIAL>"choose"       => (Tokens.CHOOSE(!pos,!pos));
<INITIAL>"nu"       => (Tokens.NU(!pos,!pos));
<INITIAL>"fst"      => (Tokens.FST(!pos,!pos));
<INITIAL>"snd"      => (Tokens.SND(!pos,!pos));

<INITIAL>"nil"      => (Tokens.NIL(!pos,!pos));
<INITIAL>"::"      => (Tokens.CONS(!pos,!pos));
<INITIAL>"hd"      => (Tokens.CAR(!pos,!pos));
<INITIAL>"tl"      => (Tokens.CDR(!pos,!pos));
<INITIAL>"print"      => (Tokens.PRINT(!pos,!pos));

<INITIAL>"raise"    => (Tokens.RAISE(!pos,!pos));
<INITIAL>"handle"   => (Tokens.HANDLER(!pos,!pos));
<INITIAL>"shift"    => (Tokens.SHIFT(!pos,!pos));
<INITIAL>"reset"   => (Tokens.RESET(!pos,!pos));
<INITIAL>"name"   => (Tokens.NAME(!pos,!pos));
<INITIAL>"unbox"   => (Tokens.UNBOX(!pos,!pos));

<INITIAL>"->"       => (Tokens.ARROWTP(!pos,!pos));
<INITIAL>"-x>"       => (Tokens.NARROWTP(!pos,!pos));
<INITIAL>"*"       => (Tokens.MULT(!pos,!pos));

<INITIAL>{alpha}({alpha}|{digit}|"_")*
               => (Tokens.IDENT(yytext,!pos,!pos));
<INITIAL>"("        => (Tokens.LPAREN(!pos,!pos));
<INITIAL>")"        => (Tokens.RPAREN(!pos,!pos));
<INITIAL>"<"        => (Tokens.LANGLE(!pos,!pos));
<INITIAL>">"        => (Tokens.RANGLE(!pos,!pos));
<INITIAL>"["        => (Tokens.LBRACK(!pos,!pos));
<INITIAL>"]"        => (Tokens.RBRACK(!pos,!pos));
<INITIAL>"+"        => (Tokens.PLUS(!pos,!pos));
<INITIAL>"-"        => (Tokens.MINUS(!pos,!pos));
<INITIAL>"="        => (Tokens.EQL(!pos,!pos));
<INITIAL>"(*"       => (YYBEGIN COMMENT;
               comment_counter := (!comment_counter) + 1;
               continue ());
<INITIAL>.      => (error ("ignoring bad character "^yytext,!pos,!pos);
             lex());

<COMMENT>"(*"           =>      (comment_counter := (!comment_counter) + 1;
                                 continue());

<COMMENT>"\n"       => (pos := (!pos) + 1; lex());

<COMMENT>"*)"           =>      (if (!comment_counter) = 1 then 
                                      YYBEGIN INITIAL else ();
                                 comment_counter := (!comment_counter) - 1;
                                 continue());
 
<COMMENT>.              =>      (continue ());
